# This file is part of GNU Anubis testsuite.
# Copyright (C) 2004-2014 The Anubis Team.
#
# This file is public domain. Do with it whatever you like to.

machine localhost login anubis password guessme
